/*$$$LICENCE_NORDIC_STANDARD<2017>$$$*/
#include "app_timer.h"
#include "nrf_queue.h"
#include "nrf_sortlist.h"
#include <stddef.h>
#include "app_timer_rtc_interface.h"
#define NRF_LOG_MODULE_NAME app_timer
#if APP_TIMER_CONFIG_LOG_ENABLED
#define NRF_LOG_LEVEL       APP_TIMER_CONFIG_LOG_LEVEL
#define NRF_LOG_INFO_COLOR  APP_TIMER_CONFIG_INFO_COLOR
#define NRF_LOG_DEBUG_COLOR APP_TIMER_CONFIG_DEBUG_COLOR
#else //APP_TIMER_CONFIG_LOG_ENABLED
#define NRF_LOG_LEVEL       0
#endif //APP_TIMER_CONFIG_LOG_ENABLED
#include "nrf_log.h"
NRF_LOG_MODULE_REGISTER();

/**
 * @brief Timer requests types.
 */
typedef enum
{
    TIMER_REQ_START,
    TIMER_REQ_STOP,
    TIMER_REQ_STOP_ALL
} app_timer_req_type_t;

/**
 * @brief Macro used by logger to stringify @ref app_timer_req_type_t.
 */
#define REQ_TYPE_STRINGIFY(_type) (((_type) == TIMER_REQ_START) ? "START" : \
                                  (((_type) == TIMER_REQ_STOP) ? "STOP" : "STOP_ALL"))
/**
 * @brief app_timer control block
 */
typedef struct
{
    nrf_sortlist_item_t         list_item;     /**< Token used by sortlist. */
    volatile uint32_t           end_val;       /**< RTC counter value when timer expires. */
    uint32_t                    repeat_period; /**< Repeat period (0 if single shot mode). */
    app_timer_timeout_handler_t handler;       /**< User handler. */
    void *                      p_context;     /**< User context. */
    volatile bool               active;        /**< Flag indicating that timer is active. */
} app_timer_cb_t;

/**
 * @brief Operation request structure.
 */
typedef struct
{
    app_timer_req_type_t type;    /**< Request type. */
    app_timer_cb_t *     p_timer; /**< Timer instance. */
} timer_req_t;

static app_timer_cb_t * volatile mp_active_timer; /**< Timer currently handled by RTC driver. */
static bool             m_global_active; /**< Flag used to globally disable all timers. */

/* Queue instance. */
NRF_QUEUE_DEF(timer_req_t, m_req_queue, APP_TIMER_CONFIG_OP_QUEUE_SIZE, NRF_QUEUE_MODE_OVERFLOW);

/* Sortlist instance. */
static bool compare_func(nrf_sortlist_item_t * p_item0, nrf_sortlist_item_t *p_item1);
NRF_SORTLIST_DEF(m_app_timer_sortlist, compare_func); /**< Sortlist used for storing queued timers. */

/**
 * @brief Function used for comparing items in sorted list.
 */
static bool compare_func(nrf_sortlist_item_t * p_item0, nrf_sortlist_item_t *p_item1)
{
    app_timer_cb_t * p0 = CONTAINER_OF(p_item0, app_timer_cb_t, list_item);
    app_timer_cb_t * p1 = CONTAINER_OF(p_item1, app_timer_cb_t, list_item);

    return (p0->end_val <= p1->end_val) ? true : false;
}

/**
 * @brief Function called on timer expiration
 *
 * Function calls user handler if timer was not stopped before. If timer is in repeated mode then
 * timer is rescheduled.
 *
 * @param p_timer Timer instance.
 *
 * @return True if reevaluation of sortlist needed (becasue it was updated).
 */
static bool timer_expire(app_timer_cb_t * p_timer)
{
    ASSERT(p_timer->handler);
    bool ret = false;
    if (m_global_active == true && p_timer != NULL && p_timer->active)
    {
#if APP_TIMER_CONFIG_USE_SCHEDULER
        app_timer_event_t timer_event;

        timer_event.timeout_handler = p_timer->handler;
        timer_event.p_context       = p_timer->p_context;
        uint32_t err_code = app_sched_event_put(&timer_event, sizeof(timer_event), app_timer_evt_get);
        APP_ERROR_CHECK(err_code);
#else
        p_timer->handler(p_timer->p_context);
#endif
        if (p_timer->repeat_period && p_timer->active)
        {
            p_timer->end_val += p_timer->repeat_period;
            nrf_sortlist_add(&m_app_timer_sortlist, &p_timer->list_item);
            ret = true;
        }
        else
        {
            p_timer->active = false;
        }
    }
    return ret;
}

/**
 * @brief Function is configuring RTC driver to trigger timeout interrupt for given timer.
 *
 * It is possible that RTC driver will indicate that timeout already occured. In that case timer
 * expires and function indicates that RTC was not configured.
 *
 * @param          p_timer Timer instance.
 * @param [in,out] p_rerun Flag indicating that sortlist reevaluation is required.
 *
 * @return True if RTC was successfully configured, false if timer already expired and RTC was not
 *         configured.
 *
 */
static bool rtc_schedule(app_timer_cb_t * p_timer, bool * p_rerun)
{
    *p_rerun = false;
    ret_code_t ret = drv_rtc_compare_set(0, p_timer->end_val);
    if (ret == NRF_SUCCESS)
    {
        return true;
    }
    else if (ret == NRF_ERROR_TIMEOUT)
    {
        *p_rerun = timer_expire(p_timer);
    }
    else
    {
        ASSERT(0);
    }
    return false;
}

static inline app_timer_cb_t * sortlist_pop(void)
{
    nrf_sortlist_item_t * p_next_item = nrf_sortlist_pop(&m_app_timer_sortlist);
    return p_next_item ? CONTAINER_OF(p_next_item, app_timer_cb_t, list_item) : NULL;
}

static inline app_timer_cb_t * sortlist_peek(void)
{
    nrf_sortlist_item_t const * p_next_item = nrf_sortlist_peek(&m_app_timer_sortlist);
    return p_next_item ? CONTAINER_OF(p_next_item, app_timer_cb_t, list_item) : NULL;
}

static inline app_timer_cb_t * sortlist_next(app_timer_cb_t * p_item)
{
    nrf_sortlist_item_t const * p_next_item = nrf_sortlist_next(&p_item->list_item);
    return p_next_item ? CONTAINER_OF(p_next_item, app_timer_cb_t, list_item) : NULL;
}
/**
 * @brief Function for deactivating all timers which are in the sorted list (active timers).
 */
static void sorted_list_stop_all(void)
{
    app_timer_cb_t * p_next;
    do
    {
        p_next = sortlist_pop();
        if (p_next)
        {
            p_next->active = false;
        }
    } while (p_next);
}

/**
 * @brief Function for handling RTC counter overflow.
 *
 * In case of overflow all active timers must have end value adjusted (reduced to 24 bit range).
 */
static void on_overflow_evt(void)
{
    NRF_LOG_INFO("Overflow EVT");
    if (mp_active_timer)
    {
        mp_active_timer->end_val &= RTC_COUNTER_COUNTER_Msk;
    }

    app_timer_cb_t *     p_next;
    p_next = sortlist_peek();
    while (p_next)
    {
        if (p_next->end_val <= RTC_COUNTER_COUNTER_Msk)
        {
            //If overflow occurs then all timers with value lower than max value expires immediately.
            UNUSED_RETURN_VALUE(timer_expire(p_next));
        }
        else
        {
            p_next->end_val &= RTC_COUNTER_COUNTER_Msk;
        }
        p_next = sortlist_next(p_next);
    }
}

/**
 * #brief Function for handling RTC compare event - active timer expiration.
 */
static void on_compare_evt(void)
{
    if (mp_active_timer)
    {
        NRF_LOG_INFO("Compare EVT active context:%x", mp_active_timer->p_context);
        UNUSED_RETURN_VALUE(timer_expire(mp_active_timer));
        mp_active_timer = NULL;
    }
    else
    {
        NRF_LOG_WARNING("Compare event but no active timer (already stopped?)");
    }
}

/**
 * @brief Function updates RTC.
 *
 * Function is called at the end of RTC interrupt when all new user request and/or timer expiration
 * occured. It configures RTC if there is any pending timer, reconfigures if the are timers with
 * shorted timeout than active one or stops RTC if there is no active timers.
 */
static void rtc_update(void)
{
    while(1)
    {
        app_timer_cb_t * p_next = sortlist_peek();
        bool rtc_reconf = false;
        if (p_next) //Candidate for active timer
        {
            if (mp_active_timer == NULL)
            {
                //There is no active timer so candidate will become active timer.
                rtc_reconf = true;
            }
            else if (mp_active_timer &&
                    (mp_active_timer->end_val > p_next->end_val))
            {
                //Candidate has shorter timeout than current active timer. Candidate will replace active timer.
                //Active timer is put back into sorted list.
                rtc_reconf = true;
                if (mp_active_timer->active)
                {
                    NRF_LOG_INFO("Preempting timer:%i", mp_active_timer->p_context);
                    nrf_sortlist_add(&m_app_timer_sortlist, &mp_active_timer->list_item);
                }
            }

            if (rtc_reconf)
            {
                bool rerun;
                p_next = sortlist_pop();
                NRF_LOG_INFO("Activation timer:%i CC:%d (%06x)", p_next->p_context, p_next->end_val, p_next->end_val);
                if (rtc_schedule(p_next, &rerun))
                {
                    if (!APP_TIMER_KEEPS_RTC_ACTIVE && (mp_active_timer == NULL))
                    {
                        drv_rtc_start();
                    }
                    mp_active_timer = p_next;

                    if (rerun == false)
                    {
                        //RTC was successfully updated and sortlist was not updated. Function can be terminated.
                        break;
                    }
                }
                else
                {
                    //If RTC driver indicated that timeout already occured a new candidate will be taken from sorted list.
                    NRF_LOG_INFO("Scheduled to late");
                }
            }
            else
            {
                //RTC will not be updated. Function can terminate.
                break;
            }
        }
        else //No candidate for active timer.
        {
            if (!APP_TIMER_KEEPS_RTC_ACTIVE && mp_active_timer == NULL)
            {
                drv_rtc_stop();
            }
            break;
        }
    }
}

/**
 * @brief Function for processing user requests.
 *
 * Function is called only in the context of RTC interrupt.
 */
static void timer_req_process(void)
{
    timer_req_t req;
    while (nrf_queue_pop(&m_req_queue, &req) == NRF_SUCCESS)
    {
        NRF_LOG_INFO("REQ:%s, context:%x", REQ_TYPE_STRINGIFY(req.type), req.p_timer->p_context);
        switch (req.type)
        {
            case TIMER_REQ_START:
                if (!req.p_timer->active)
                {
                    req.p_timer->active = true;
                    if (req.p_timer->end_val - drv_rtc_counter_get() > APP_TIMER_RTC_MAX_VALUE)
                    {
                        //A little trick to handle case when timer was scheduled just before overflow.
                        req.p_timer->end_val &= RTC_COUNTER_COUNTER_Msk;
                    }
                    nrf_sortlist_add(&m_app_timer_sortlist, &(req.p_timer->list_item));
                }
                break;
            case TIMER_REQ_STOP:
                if (req.p_timer == mp_active_timer)
                {
                    mp_active_timer = NULL;
                }
                UNUSED_RETURN_VALUE(nrf_sortlist_remove(&m_app_timer_sortlist, &(req.p_timer->list_item)));
                break;
            case TIMER_REQ_STOP_ALL:
                sorted_list_stop_all();
                m_global_active = true;
                break;
            default:
                break;
        }
    }
}

/**
 * @brief Function for handling RTC interrupt.
 */
static void rtc_irq(void)
{
    if (drv_rtc_overflow_pending())
    {
        on_overflow_evt();
    }
    if (drv_rtc_compare_pending(0))
    {
        on_compare_evt();
    }
    timer_req_process();
    rtc_update();
}

/**
 * @brief Function for triggering processing user requests.
 *
 * @note All user requests are processed in a single context - RTC interrupt.
 */
static inline void timer_request_proc_trigger(void)
{
    drv_rtc_irq_trigger();
}

/**
 * @brief Function for putting user request into the request queue
 */
static ret_code_t timer_req_schedule(timer_req_t * p_req)
{
    ret_code_t ret;
    ret = nrf_queue_push(&m_req_queue, p_req);
    if (ret == NRF_SUCCESS)
    {
        timer_request_proc_trigger();
    }
    return ret;
}

ret_code_t app_timer_init(void)
{
    ret_code_t err_code;
    drv_rtc_config_t config = {
            .prescaler   = APP_TIMER_CONFIG_RTC_FREQUENCY,
           .irq_priority = APP_TIMER_CONFIG_IRQ_PRIORITY,
           .irq_func     = rtc_irq
    };

    err_code = drv_rtc_init(&config);
    drv_rtc_overflow_enable(true);
    if (APP_TIMER_KEEPS_RTC_ACTIVE)
    {
        drv_rtc_start();
    }

    m_global_active = true;
    return err_code;
}

ret_code_t app_timer_create(app_timer_id_t const *      p_timer_id,
                            app_timer_mode_t            mode,
                            app_timer_timeout_handler_t timeout_handler)
{

    if (timeout_handler == NULL)
    {
        return NRF_ERROR_INVALID_PARAM;
    }

    app_timer_cb_t * p_t = (app_timer_cb_t *) *p_timer_id;
    p_t->handler = timeout_handler;
    p_t->repeat_period = (mode == APP_TIMER_MODE_REPEATED) ? 1 : 0;
    return NRF_SUCCESS;
}

ret_code_t app_timer_start(app_timer_t * p_timer, uint32_t timeout_ticks, void * p_context)
{
    app_timer_cb_t * p_t = (app_timer_cb_t *) p_timer;

    if (timeout_ticks > APP_TIMER_RTC_MAX_VALUE)
    {
        return NRF_ERROR_INVALID_PARAM;
    }
    p_t->p_context = p_context;
    p_t->end_val = drv_rtc_counter_get() + timeout_ticks;

    if (p_t->repeat_period)
    {
        p_t->repeat_period = timeout_ticks;
    }

    timer_req_t req = {
        .type = TIMER_REQ_START,
        .p_timer = p_t
    };

    return timer_req_schedule(&req);
}


ret_code_t app_timer_stop(app_timer_t * p_timer)
{
    app_timer_cb_t * p_t = (app_timer_cb_t *) p_timer;
    p_t->active = false;

    timer_req_t req = {
        .type = TIMER_REQ_STOP,
        .p_timer = (app_timer_cb_t *)p_timer
    };

    return timer_req_schedule(&req);
}

ret_code_t app_timer_stop_all(void)
{
    //block timer globally
    m_global_active = false;
    timer_req_t req = {
        .type = TIMER_REQ_STOP_ALL
    };

    return timer_req_schedule(&req);
}

uint8_t app_timer_op_queue_utilization_get(void)
{
    return (uint8_t)nrf_queue_utilization_get(&m_req_queue);
}

uint32_t app_timer_cnt_diff_compute(uint32_t   ticks_to,
                                    uint32_t   ticks_from)
{
    return ((ticks_to - ticks_from) & RTC_COUNTER_COUNTER_Msk);
}

uint32_t app_timer_cnt_get(void)
{
    return drv_rtc_counter_get();
}

void app_timer_pause(void)
{
    drv_rtc_stop();
}

void app_timer_resume(void)
{
    drv_rtc_start();
}
