/*$$$LICENCE_NORDIC_STANDARD<2017>$$$*/
#include "app_timer_rtc_interface.h"
#include "sdk_errors.h"
#include "nrf_drv_common.h"
#include "nrf_drv_clock.h"
#include "nrf_delay.h"
#include "nrfx.h"
#include <hal/nrf_rtc.h>
#define NRF_LOG_MODULE_NAME app_timer_rtc
#include "nrf_log.h"
NRF_LOG_MODULE_REGISTER();

static rtr_irq_t m_irq_func;

ret_code_t drv_rtc_init(drv_rtc_config_t const * p_config)
{
    ret_code_t err_code = nrf_drv_clock_init();

    if (err_code != NRF_SUCCESS && err_code != NRFX_ERROR_ALREADY_INITIALIZED)
    {
        return err_code;
    }

    nrf_drv_clock_lfclk_request(NULL);
    m_irq_func = p_config->irq_func;
    nrf_rtc_prescaler_set(NRF_RTC1, p_config->prescaler);
    NRFX_IRQ_PRIORITY_SET(RTC1_IRQn, p_config->irq_priority);
    NRFX_IRQ_ENABLE(RTC1_IRQn);
    return NRF_SUCCESS;
}

void drv_rtc_uninit(void)
{
    NRFX_IRQ_DISABLE(RTC1_IRQn);
    drv_rtc_stop();
    nrf_drv_clock_lfclk_release();
}

void drv_rtc_start()
{
    nrf_rtc_task_trigger(NRF_RTC1, NRF_RTC_TASK_START);
}

void drv_rtc_stop(void)
{
    nrf_rtc_task_trigger(NRF_RTC1, NRF_RTC_TASK_STOP);
}

ret_code_t drv_rtc_compare_set(uint32_t cc, uint32_t abs_value)
{
    abs_value &=RTC_COUNTER_COUNTER_Msk;
    nrf_rtc_int_t cc_int_mask = (nrf_rtc_int_t)(NRF_RTC_INT_COMPARE0_MASK << cc);
    nrf_rtc_event_t cc_evt = (nrf_rtc_event_t)offsetof(NRF_RTC_Type,EVENTS_COMPARE[cc]);
    nrf_rtc_int_disable(NRF_RTC1, cc_int_mask);
    nrf_rtc_event_disable(NRF_RTC1, cc_int_mask);
    nrf_rtc_event_clear(NRF_RTC1, cc_evt);
    nrf_rtc_cc_set(NRF_RTC1,cc,abs_value);
    nrf_rtc_event_enable(NRF_RTC1, cc_int_mask);

    int32_t diff = abs_value - drv_rtc_counter_get();
    diff = diff < 0 ? -diff : diff;
    if (diff < APP_TIMER_RTC_SAFE_WINDOW)
    {
        nrf_delay_us(33);
        uint32_t counter = nrf_rtc_counter_get(NRF_RTC1);
        int32_t diff2 = counter - abs_value;
        if (counter < abs_value)
        {
            diff2 += RTC_COUNTER_COUNTER_Msk;
        }
        if (drv_rtc_compare_pending(cc) ||
                ((diff2 >= -1) && (diff2 < APP_TIMER_RTC_SAFE_WINDOW)))
        {
            NRF_LOG_WARNING("Compare set timeout cc_pending:%d, diff:%d, counter:%x, abs_val:%x", drv_rtc_compare_pending(cc), diff2, counter, abs_value);
            return NRF_ERROR_TIMEOUT;
        }

        nrf_rtc_int_enable(NRF_RTC1, cc_int_mask);
        if (nrf_rtc_event_pending(NRF_RTC1,cc_evt))
        {
            nrf_rtc_int_disable(NRF_RTC1, cc_int_mask);
            nrf_rtc_event_disable(NRF_RTC1, cc_int_mask);
            nrf_rtc_event_clear(NRF_RTC1, cc_evt);
            NRF_LOG_WARNING("Compare set timeout, event pending");
            return NRF_ERROR_TIMEOUT;
        }
    }
    else
    {
        nrf_rtc_int_enable(NRF_RTC1, cc_int_mask);
    }
    return NRF_SUCCESS;
}

static void evt_enable(uint32_t mask, bool irq_enable)
{
    nrf_rtc_event_enable(NRF_RTC1, mask);
    if (irq_enable)
    {
        nrf_rtc_int_enable(NRF_RTC1, mask);
    }
}

static void evt_disable(uint32_t mask)
{
    nrf_rtc_event_disable(NRF_RTC1, mask);
    nrf_rtc_int_disable(NRF_RTC1, mask);
}

static bool evt_pending(nrf_rtc_event_t event)
{
    bool            ret   = false;
    if (nrf_rtc_event_pending(NRF_RTC1,event))
    {
        nrf_rtc_event_clear(NRF_RTC1,event);
        ret = true;
    }
    return ret;
}


void drv_rtc_overflow_enable(bool irq_enable)
{
    evt_enable(NRF_RTC_INT_OVERFLOW_MASK, irq_enable);
}

void drv_rtc_overflow_disable(void)
{
    evt_disable(NRF_RTC_INT_OVERFLOW_MASK);
}

bool drv_rtc_overflow_pending(void)
{
    return evt_pending(NRF_RTC_EVENT_OVERFLOW);
}

void drv_rtc_tick_enable(bool irq_enable)
{
    evt_enable(NRF_RTC_INT_TICK_MASK, irq_enable);
}

void drv_rtc_tick_disable(void)
{
    evt_disable(NRF_RTC_INT_TICK_MASK);
}

bool drv_rtc_tick_pending(void)
{
    return evt_pending(NRF_RTC_EVENT_TICK);
}

void drv_rtc_compare_enable(uint32_t cc, bool irq_enable)
{
    evt_enable(NRF_RTC_INT_COMPARE0_MASK << cc, irq_enable);
}

void drv_rtc_compare_disable(uint32_t cc)
{
    evt_disable(NRF_RTC_INT_COMPARE0_MASK << cc);
}


bool drv_rtc_compare_pending(uint32_t cc)
{
    nrf_rtc_event_t cc_evt = (nrf_rtc_event_t)offsetof(NRF_RTC_Type,EVENTS_COMPARE[cc]);
    return evt_pending(cc_evt);
}

uint32_t drv_rtc_counter_get(void)
{
    return nrf_rtc_counter_get(NRF_RTC1);
}

void drv_rtc_irq_trigger(void)
{
    NVIC_SetPendingIRQ(RTC1_IRQn);
}

void RTC1_IRQHandler(void)
{
    m_irq_func();
}
