/*$$$LICENCE_NORDIC_STANDARD<2017>$$$*/
#ifndef APP_TIMER_RTC_INTERFACE_H
#define APP_TIMER_RTC_INTERFACE_H
#include "sdk_errors.h"
#include <stdint.h>
#include <stdbool.h>

/**
 * @brief RTC safe window.
 * Maximum possible relative value is limited by safe window to detect cases when requested
 * compare event has already occured.
 */
#define APP_TIMER_RTC_SAFE_WINDOW 1000

/**
 * @brief RTC maximum counter value.
 */
#define APP_TIMER_RTC_MAX_CNT     0xFFFFFF

/**
 * @brief Maximum relative timeout which can be set.
 */
#define APP_TIMER_RTC_MAX_VALUE   (APP_TIMER_RTC_MAX_CNT - APP_TIMER_RTC_SAFE_WINDOW)

/** @brief RTC interrupt function prototype. */
typedef void (*rtr_irq_t)(void);

/**
 * @brief Configuration structure of RTC driver.
 */
typedef struct
{
    uint16_t prescaler;
    uint8_t  irq_priority;
    rtr_irq_t irq_func;
} drv_rtc_config_t;

/**
 * @brief Function for initializing RTC driver
 *
 * @param p_config Configuration.
 * @return NRF_SUCCESS or error code.
 */
ret_code_t drv_rtc_init(drv_rtc_config_t const * p_config);

/**
 * @brief Function for starting RTC clock.
 */
void drv_rtc_start(void);

/**
 * @brief Function for stopping RTC clock.
 */
void drv_rtc_stop(void);

/**
 * @brief Function for configuring compare channel.
 *
 * @param cc        Compare channel index.
 * @param abs_value Absolute value to be set in the compare register.
 *
 * @retval NRF_SUCCESS       Successful operation
 * @retval NRF_ERROR_TIMEOUT Compare value already reached.
 */
ret_code_t drv_rtc_compare_set(uint32_t cc, uint32_t abs_value);

/**
 * @brief Function for getting current value of RTC counter.
 *
 * @return Counter value.
 */
uint32_t drv_rtc_counter_get(void);

/**
 * @brief Function for enabling overflow event and interrupt.
 *
 * @param irq_enable True to enable interrupt.
 */
void drv_rtc_overflow_enable(bool irq_enable);

/**
 * @brief Function for diabling overflow event and interrupt.
 */
void drv_rtc_overflow_disable(void);

/**
 * @brief Function for checking if overflow event has occured.
 *
 * @note Event is cleared after reading.
 *
 * @return True if interrupt pending, false otherwise.
 */
bool drv_rtc_overflow_pending(void);

void drv_rtc_tick_enable(bool irq_enable);
void drv_rtc_tick_disable(void);
bool drv_rtc_tick_pending(void);

void drv_rtc_compare_enable(uint32_t cc, bool irq_enable);
void drv_rtc_compare_disable(uint32_t cc);
bool drv_rtc_compare_pending(uint32_t cc);

void drv_rtc_irq_trigger(void);

#endif //APP_TIMER_RTC_INTERFACE_H
